/*
  *************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *  
 *************************************************************************
*/

package org.eclipse.birt.data.engine.aggregation;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.birt.data.engine.aggregation.rank.TotalIsBottomN;
import org.eclipse.birt.data.engine.aggregation.rank.TotalIsBottomNPercent;
import org.eclipse.birt.data.engine.aggregation.rank.TotalIsTopN;
import org.eclipse.birt.data.engine.aggregation.rank.TotalIsTopNPercent;
import org.eclipse.birt.data.engine.aggregation.rank.TotalPercentRank;
import org.eclipse.birt.data.engine.aggregation.rank.TotalPercentSum;
import org.eclipse.birt.data.engine.aggregation.rank.TotalPercentile;
import org.eclipse.birt.data.engine.aggregation.rank.TotalQuartile;
import org.eclipse.birt.data.engine.aggregation.rank.TotalRank;
import org.eclipse.birt.data.engine.api.aggregation.IAggregation;
import org.eclipse.birt.data.engine.api.aggregation.IBuildInAggregation;

/**
 * This class provides instances of all built-in BIRT aggregations.
 */
public class BuiltInAggregationFactory implements IBuildInAggregation 
{
    private static BuiltInAggregationFactory instance = null;
    
    private static Map aggregations = new HashMap();
    
    static
	{
    	aggregations.put( TOTAL_COUNT_FUNC, new TotalCount( ) );
		aggregations.put( TOTAL_SUM_FUNC, new TotalSum( ) );
		aggregations.put( TOTAL_MAX_FUNC, new TotalMax( ) );
		aggregations.put( TOTAL_MIN_FUNC, new TotalMin( ) );
		aggregations.put( TOTAL_AVE_FUNC, new TotalAve( ) );
		aggregations.put( TOTAL_WEIGHTEDAVE_FUNC, new TotalWeightedAve( ) );
		aggregations.put( TOTAL_MOVINGAVE_FUNC, new TotalMovingAve( ) );
		aggregations.put( TOTAL_MEDIAN_FUNC, new TotalMedian( ) );
		aggregations.put( TOTAL_MODE_FUNC, new TotalMode( ) );
		aggregations.put( TOTAL_STDDEV_FUNC, new TotalStdDev( ) );
		aggregations.put( TOTAL_VARIANCE_FUNC, new TotalVariance( ) );
		aggregations.put( TOTAL_FIRST_FUNC, new TotalFirst( ) );
		aggregations.put( TOTAL_LAST_FUNC, new TotalLast( ) );
		aggregations.put( TOTAL_RUNNINGSUM_FUNC, new TotalRunningSum( ) );
		aggregations.put( TOTAL_IRR_FUNC, new TotalIrr( ) );
		aggregations.put( TOTAL_MIRR_FUNC, new TotalMirr( ) );
		aggregations.put( TOTAL_NPV_FUNC, new TotalNpv( ) );
		aggregations.put( TOTAL_RUNNINGNPV_FUNC, new TotalRunningNpv( ) );
		aggregations.put( TOTAL_COUNTDISTINCT_FUNC, new TotalCountDistinct( ) );
		aggregations.put( TOTAL_RANK_FUNC, new TotalRank( ) );
		aggregations.put( TOTAL_TOP_N_FUNC, new TotalIsTopN( ) );
		aggregations.put( TOTAL_TOP_PERCENT_FUNC, new TotalIsTopNPercent( ) );
		aggregations.put( TOTAL_BOTTOM_N_FUNC, new TotalIsBottomN( ) );
		aggregations.put( TOTAL_BOTTOM_PERCENT_FUNC, new TotalIsBottomNPercent( ) );
		aggregations.put( TOTAL_PERCENT_RANK_FUNC, new TotalPercentRank( ) );
		aggregations.put( TOTAL_PERCENTILE_FUNC, new TotalPercentile( ) );
		aggregations.put( TOTAL_QUARTILE_FUNC, new TotalQuartile( ) );
		aggregations.put( TOTAL_PERCENTSUM_FUNC, new TotalPercentSum( ) );
		aggregations.put( TOTAL_RUNNINGCOUNT_FUNC, new TotalRunningCount( ) );
    }
   
    private static String[] aggNames = {
    	TOTAL_COUNT_FUNC,
        TOTAL_SUM_FUNC,
        TOTAL_MAX_FUNC,
        TOTAL_MIN_FUNC,
        TOTAL_AVE_FUNC,
        TOTAL_WEIGHTEDAVE_FUNC,
        TOTAL_MOVINGAVE_FUNC,
        TOTAL_MEDIAN_FUNC,
        TOTAL_MODE_FUNC,
        TOTAL_STDDEV_FUNC,
        TOTAL_VARIANCE_FUNC,
        TOTAL_FIRST_FUNC,
        TOTAL_LAST_FUNC,
        TOTAL_RUNNINGSUM_FUNC,
        
        TOTAL_IRR_FUNC,
        TOTAL_MIRR_FUNC,
        TOTAL_NPV_FUNC,
        TOTAL_RUNNINGNPV_FUNC,
        TOTAL_COUNTDISTINCT_FUNC,
        
        TOTAL_RANK_FUNC,
        TOTAL_TOP_N_FUNC,
        TOTAL_TOP_PERCENT_FUNC,
        TOTAL_BOTTOM_N_FUNC,
        TOTAL_BOTTOM_PERCENT_FUNC,
        
        TOTAL_PERCENT_RANK_FUNC,
        TOTAL_PERCENTILE_FUNC,
        TOTAL_QUARTILE_FUNC,
        TOTAL_PERCENTSUM_FUNC,
        
        TOTAL_RUNNINGCOUNT_FUNC
    }; 
    
    private BuiltInAggregationFactory( )
	{
	};
    
    /**
     * @return
     */
    public static BuiltInAggregationFactory getInstance( )
	{
		if ( instance == null )
		{
			synchronized ( BuiltInAggregationFactory.class )
			{
				if ( instance == null )
					instance = new BuiltInAggregationFactory( );
			}
		}

		return instance;
	}
    

                               
    /**
     * Gets an instance of the built-in Aggregation identified by name.
     * @param name Name (e.g., "SUM", "AVG", etc.) of the built-in aggregation. This name is
     * case insensitive
     * @return Instance of the requested Aggregation, or null if no such name is found. 
     */
    public IAggregation getAggregation( String name )
    {
    	return (IAggregation) aggregations.get( name.toUpperCase( ) );
    }

	public String[] getAggregationNames( )
	{
		return aggNames;
	}
}


